/**
 * File:
 *  anim.c
 *
 * Notes:
 *  This file contains the implementation of the animation object.  This can be 
 * used to create a animation sequences on LED's.
 *
 * outputScheduleDef:
 *  bPin             The pin to send to the port
 *  uintPreDelay     The count to wait before applying
 *  uintPostDelay    The count to wait after applying
 *  faderDef fader   see fader.h
 *
 * animDef:
 * Bit Flags:
 *  uLoop            Flag to loop
 *  uBounce          Flag to indicate bounce required
 *  uDir             Flag to indicate direction 0=Up, 1=Dwn
 *  uStop            Flag to indicate animation complete
 * Members:
 *  pPort            The port the output is associated with
 *  bTotalSteps      The number of steps in the animation
 *  parySeq          The sequence associated with this anim
 *  intIndex;        Place in sequence
 *  bStage;          Stage in current animation
 *  uintCtr;         Current delay value
 *  pCurrent;        Link to current animation from links
 *  pHead;           Link to first node in animation
 *  pTail;           Link to last node in animation
 *  pPrev;           Link to previous node in animation
 *  pNext;           Link to next node in animation
 *
 * Functions:
 *  playAnimation    Plays an animation sequence
 *
 * History:
 *  23/05/2012 Created by Simon Platten
 */
#include "anim.h"
/**
 * Function:
 *  initAnim
 *
 * Parameters:
 *  pAnim, the animation to initialise
 *  intSteps, the number of steps in the fading sequence returned from
 *            createSchedule
 * Returns:
 *  none
 */
void initAnim(animDef* pAnim, int intSteps) {
  outputScheduleDef* pStep;
  int i;
  for( i=0; i<pAnim->bTotalSteps; i++ ) {
    pStep = &pAnim->parySeq[i];
    createFader(&pStep->fader, pAnim->pPort, pStep->bPin,
                (uint)intSteps, TRUE, FALSE, FALSE, FALSE, TRUE, NULL);
  }
}
/**
 * Function:
 *  joinAnims
 *
 * Parameters:
 *  pDst, the master animation to join with
 *  pSrc, the animation to join from
 *
 * Returns:
 *  none
 */
void joinAnims(animDef* pDst, animDef* pSrc) {
  if ( pDst == NULL || pSrc == NULL ) {
// Invalid parameters abort
    return;
  }
  if ( pDst->pNext == NULL ) {
// First join
    pSrc->pPrev = pDst;
    pDst->pNext = pSrc;
  } else {
// Find last join
    animDef* pNext;
    for( pNext=pDst; pNext->pNext!=NULL; pNext=pNext->pNext ) {
         pNext->pTail = pSrc;
    }
    pSrc->pPrev = pNext;
    pNext->pNext = pNext->pTail = pSrc;
  }
  pDst->pHead = pSrc->pHead = pDst;
  pDst->pTail = pSrc->pTail = pSrc;
// Initialise indexes based on direction
  pSrc->intIndex = (pSrc->flags.uDir == 0) ? 0 : pSrc->bTotalSteps - 1;
  pDst->intIndex = (pDst->flags.uDir == 0) ? 0 : pDst->bTotalSteps - 1;
  
  if ( pDst->pHead->flags.uDir == 0 ) {
// Make sure the current node starts at the first node
    pDst->pHead->pCurrent = pDst->pHead;
  } else {
// Make sure the current node starts at the last node
    pDst->pHead->pCurrent = pDst->pTail;
  }
}
/**
 * Function:
 *  playAnimation
 *
 * Parameters:
 *  pAnim, pointer to animation and sequence
 *
 * Returns:
 *  none
 */
void playAnimation(animDef* pAnim) {
  outputScheduleDef* pCurSeq;
  animDef* pCurAnim;

  if ( pAnim->flags.uStop ) {
// Animation complete and stopped, do nothing
     return;
  }
  if ( pAnim->pCurrent == NULL ) {
// No current animation use the this animation as the reference
    pAnim->pCurrent = pAnim;
  }
  pCurAnim = pAnim->pCurrent;
// Get the current step in the animation
  if ( pCurAnim->flags.uDir == 0 ) {
// Direction, up through schedule
    if ( pCurAnim->intIndex >= (int)pCurAnim->bTotalSteps ) {
      if ( pCurAnim->flags.uBounce == 1 ) {
        pCurAnim->flags.uDir = ~pCurAnim->flags.uDir;
        pCurAnim->intIndex = (int)(pCurAnim->bTotalSteps - 1);
      } else {
        if ( !pCurAnim->flags.uLoop ) {
// No loop, flag that animation is complete and abort
          pCurAnim->flags.uStop = 1;
          return;
        }
// Reset index and start over
        pCurAnim->intIndex = 0;
      }
// Is there another animation sequence joined to this ?
      if ( pCurAnim->pNext ) {
// Yes, change the current animation to the next sequence
        pCurAnim = pCurAnim->pNext;
      } else if ( !pCurAnim->flags.uBounce && pCurAnim->pHead ) {
        pCurAnim = pCurAnim->pHead;
      }
      if ( pCurAnim->pHead ) {
        pCurAnim->pHead->pCurrent = pCurAnim;
      }
    }
  } else {
// Direction, down through schedule
    if ( pCurAnim->intIndex < 0 ) {
      if ( pCurAnim->flags.uBounce == 1 ) {
        pCurAnim->flags.uDir = ~pCurAnim->flags.uDir;
        pCurAnim->intIndex = 0;
      } else {
        if ( !pCurAnim->flags.uLoop ) {
// No loop, flag that animation is complete and abort
          pCurAnim->flags.uStop = 1;
          return;
        }
// Reset index and start over
        pCurAnim->intIndex = (int)(pCurAnim->bTotalSteps - 1);
      }
// Is there another animation sequence joined to this ?
      if ( pCurAnim->pPrev ) {
// Yes, change the current animation to the previous sequence
        pCurAnim = pCurAnim->pPrev;
      } else if ( !pCurAnim->flags.uBounce && pCurAnim->pTail ) {
        pCurAnim = pCurAnim->pTail;
      }
      if ( pCurAnim->pHead ) {
        pCurAnim->pHead->pCurrent = pCurAnim;
      }
    }
  }
  pCurSeq = pCurAnim->parySeq;
  pCurSeq = &pCurSeq[pCurAnim->intIndex];

  switch((int)pCurAnim->bStage) {
  case 0:
// Need to initialise this step in the animation
    if ( pCurSeq->uintPreDelay == 0 ) {
// No pre-action delay, go straight to applying mask
      pCurAnim->bStage = 2;
      break;
    }
    pCurAnim->uintCtr = pCurSeq->uintPreDelay;
    pCurAnim->bStage = 1;
    break;
  case 1:
// Pre-action delay
    if ( --pCurAnim->uintCtr == 0 ) {
      pCurAnim->bStage = 2;
    } else {
      break;
    }
  case 2:
// Apply mask
    *pCurAnim->pPort = (1 << pCurSeq->bPin);
// Start the fader
    pCurSeq->fader->intStep = pCurSeq->fader->uintSteps - 1;
    pCurSeq->fader->uStop = 0;

    if ( pCurSeq->uintPostDelay > 0 ) {
      pCurAnim->uintCtr = pCurSeq->uintPostDelay;
      pCurAnim->bStage = 3;
    } else {
      pCurAnim->bStage = 4;
      break;
    }
  case 3:
// Post action delay
    if ( --pCurAnim->uintCtr == 0 ) {
      pCurAnim->bStage = 4;
    }
    break;
  case 4:
// Next step in sequence
    if ( pCurAnim->flags.uDir == 0 ) {
      pCurAnim->intIndex++;
    } else {
      pCurAnim->intIndex--;
    }
    pCurAnim->bStage = 0;
    break;
  }
  serviceFaders();
}