/**
 * File:
 *  fader.c
 *
 * Notes:
 *  This file contains the implementation of the fader object.  This can be used
 * to create a fade in or fade out effect on an output pin.
 *
 * faderDef:
 * Bit Flags:
 *  uBounce,           0 = Off, 1 = Bounce back to initial state
 *  uDirection,        0 = Fade In, 1 = Fade Out
 *  uLoop,             0 = Off, 1 = Continous loop
 *  uPin,              0 to 7, the output pin
 *  uState,            0 = On, 1 = Off
 *  uStop,             0 = Running, 1 = Stopped
 * Members:
 *  pPort              The port the output is associated with
 *  uintSteps          The number of steps in the schedule
 *  intStep            The current step in the schedule
 *  uintLastStep       The last step performed
 *  uintCtr            The current count of the action in the sch.
 *  uintRepeat         The number of times to repeat the step
 *  pData              Data to pass onto complete call-back
 *  pCB                Call-back to call when complete
 *  pSchItem           The current item in the schedule
 *  pNext              The next node in the linked list
 *
 * Functions:
 *  createFader        creates a fader initialising flags to specified state
 *  createSchedule     creates a fading schedule
 *  serviceFaders      services all configured faders
 *
 * History:
 *  07/05/2012 Created by Simon Platten
 *  13/05/2012 Added createSchedule
 *  14/05/2012 Modified createSchedule to use a static array for the schedule
 *             Limitation, you can have only one schedule!
 */
#include "fader.h"

static scheduleDef arySchedule[100];
static faderDef* pFaderLL = NULL;
/**
 * Function:
 *  getStepFromSchedule
 *
 * Parameters:
 *  pFader, the fader to modify
 *
 * Returns:
 *  none
 */
#define getStepFromSchedule(pFader) { \
  pFader->pSchItem = &arySchedule[pFader->intStep]; \
  pFader->uintCtr = pFader->pSchItem->aryuintStates[pFader->uState]; \
  if ( pFader->intLastStep != pFader->intStep ) { \
    pFader->intLastStep = pFader->intStep; \
    pFader->uintRepeat = pFader->pSchItem->uintRepeat; \
  } \
}
/**
 * Function:
 *  createFader
 *
 * Parameters:
 *  pFader, a pointer to an instance of faderDef to configure
 *  uPin, the output pin number, 0 to N.
 *  uintSteps, optional, the number of steps in the schedule
 *  blnDir, FALSE = fade in, TRUE = fade out
 *  blnBoune, FALSE = no bounce, TRUE = bounce
 *  blnLoop, FALSE = no, TRUE = loop
 *  blnState, initial state, TRUE = On, FALSE = Off
 *  blnStopped, FALSE = start immediately, TRUE = stopped state
 *  pCB, optional call-back to call when complete
 *
 * Returns:
 *  true if ok, false if error
 */
boolean createFader(faderDef* pFader,
                    byte* pPort,
                    byte uPin,
                    uint uintSteps,
                    boolean blnDir,
                    boolean blnBounce,
                    boolean blnLoop,
                    boolean blnState,
                    boolean blnStopped,
                    void (*pCB)()) {
  if ( pFader == NULL || pPort == NULL || uPin > 8 || uintSteps == 0 ) {
    return FALSE;
  }
  memset(pFader, 0, sizeof(faderDef));
  pFader->pPort      = pPort;
  pFader->uPin       = uPin;
  pFader->uDirection = (blnDir    == TRUE) ? 1 : 0;
  pFader->uBounce    = (blnBounce == TRUE) ? 1 : 0;
  pFader->uLoop      = (blnLoop   == TRUE) ? 1 : 0;
  pFader->uState     = (blnState  == TRUE) ? 1 : 0;
  pFader->uStop      = blnStopped;
  pFader->pCB        = pCB;
  pFader->uintSteps  = uintSteps;

  if ( pFader->uDirection == 1 ) {
    pFader->intStep = pFader->uintSteps - 1;
  }
// Make sure the last step is different from step
  pFader->intLastStep = pFader->intStep + 1;
// Set-up the schedule item
  getStepFromSchedule(pFader);
// Add the fader to the linked list
  if ( pFaderLL == NULL ) {
    pFaderLL = pFader;
  } else {
// Find the end of the list
    faderDef* pTail;
    for( pTail=pFaderLL; pTail->pNext!=NULL; pTail=pTail->pNext ) {
    }
    pTail->pNext = pFader;
  }
  return TRUE;
}
/**
 * Function:
 *  createSchedule
 *
 * Parameters:
 *  uintInc, the amount to increment in each step
 *  uintRepeat, the number of times to repeat each item in the schedule
 *
 * Returns:
 *  The number of steps in the schedule
 */
int createSchedule(uint uintInc, uint uintRepeat) {
  int intSteps, intStep, intRemaining;

  if ( uintInc == 0 ) {
// Invalid parameters passed, abort!
    return 0;
  }
// Start at 100% and decrement until <= 0
  intSteps = 0;
  intRemaining = 100;
  do {
    intSteps++;
    intRemaining -= uintInc;
  } while( intRemaining > 0 );
// Allocate memory for the schedule
  memset(arySchedule, 0, sizeof(arySchedule));
  for( intStep=0; intStep<intSteps; intStep++ ) {
    uint uintOn;
    uintOn = (uint)intStep * (uint)uintInc;
    arySchedule[intStep].aryuintStates[0] = uintOn;
    arySchedule[intStep].aryuintStates[1] = max(0, 100 - uintOn);
    arySchedule[intStep].uintRepeat = uintRepeat;
  }
  return intSteps;
}
/**
 * Function:
 *  serviceFaders
 *
 * Parameters:
 *  none
 *
 * Returns:
 *  none
 */
void serviceFaders(void) {
  faderDef* pFader;
  int intBit;
  for( pFader=pFaderLL; pFader!=NULL; pFader=pFader->pNext ) {
    if ( pFader->uStop ) {
// This fader isn't active, skip it!
      continue;
    }
// Get the correct bit according to the pin
    intBit = 1 << pFader->uPin;

    if ( pFader->uState == 0 ) {
// On
      *pFader->pPort |= intBit;
    } else {
// Off
      *pFader->pPort &= 0xffff ^ intBit;
    }
    if ( pFader->uintCtr == 0 ) {
      if ( pFader->uState == 1 ) {
// At the 2nd state in the current item in the schedule, are they're any more
// repeat actions
        if ( pFader->uintRepeat == 0 ) {
// No more repeats, time to more onto the next item in the schedule
          if ( pFader->uDirection == 0 ) {
// Moving up from the beginning of the schedule
            if ( ++pFader->intStep >= pFader->uintSteps ) {
// We are at the end of the schedule, what action do we take?
              if ( pFader->uLoop ) {
// Loop, reset to beginning of the schedule
                pFader->intStep = 0;
              } else
              if ( pFader->uBounce ) {
// Bound, change direction
                pFader->uDirection = 1;
                pFader->intStep = pFader->uintSteps - 1;
              } else {
// No further action required stop!
                pFader->uStop = 1;
                return;
              }
            }
          } else {
// Moving down from the end of the schedule
            if ( --pFader->intStep <= 0 ) {
// We are at the beginning of the schedule, what action do we take?
              if ( pFader->uLoop ) {
// Loop, reset to end of the schedule
                pFader->intStep = pFader->uintSteps - 1;
              } else if ( pFader->uBounce ) {
// Bound, change direction
                pFader->uDirection = 0;
                pFader->intStep = 0;
              } else {
// No further action required stop!
                pFader->uStop = 1;
                return;
              }
            }
          }
        } else {
// Still repeats to do, decrement repeat counter and go again!
          pFader->uintRepeat--;
        }
      }
// Toggle state
      pFader->uState = ~pFader->uState;
// Get the next item from the schedule
      getStepFromSchedule(pFader);

      if ( pFader->pCB != NULL ) {
        (*pFader->pCB)();
      }
    } else {
// Decrement the counter
      pFader->uintCtr--;
    }
  }
}